"""
Where
====================================
Core functions to construct conditional statements and manage them
"""
from enum import Enum
from typing import List


class WhereStatementKeys(Enum):
    """Enum of different operators in a where statement"""
    EQUAL = 'EQUAL'
    NO_EQUAL = 'NOEQUAL'
    MORE = 'MORE'
    LESS = 'LESS'
    MORE_EQUAL = 'MOREEQUAL'
    LESSEQUAL = 'LESSEQUAL'


class WhereAttributesKeys(Enum):
    """Enum for the conditional keys"""
    OPERATION = 'OPERATION'
    FIELD = 'FIELD'
    VALUE = 'VALUE'


WHERE_STMTS = {
    "==": WhereStatementKeys.EQUAL.value,
    "!=": WhereStatementKeys.NO_EQUAL.value,
    "<": WhereStatementKeys.LESS.value,
    ">": WhereStatementKeys.MORE.value,
    "<=": WhereStatementKeys.LESSEQUAL.value,
    ">=": WhereStatementKeys.MORE_EQUAL.value
}

WHERE_STMTS_REVERSE = {v: k for k, v in WHERE_STMTS.items()}


# FIXME: Need an AST for where conditions also check the order
def parse_where(where: str) -> List[dict]:
    """Construct the conditional statement.

    Build a list of conditional statements with a specified structure.

    Parameters
    ----------
    where : str
        Conditional statement.

    Returns
    -------
    List[dict]
        List of where statements structured in a specified way.
    """
    if where is None or where == ():
        return []
    where_clauses = []
    w_clauses = where.split(",")
    for w in w_clauses:
        wh = w.split()

        if len(wh) == 3:
            try:
                stmt = {WhereAttributesKeys.OPERATION.value: WHERE_STMTS[wh[1]],
                        WhereAttributesKeys.FIELD.value: wh[0],
                        WhereAttributesKeys.VALUE.value: wh[2]}
            except KeyError:
                raise ValueError(f"Unknown \"where\" syntax.")
            where_clauses.append(stmt)
        else:
            raise ValueError(f"Unknown where syntax.")
    return where_clauses


def skip(row: dict, where: List[dict]) -> bool:
    """Check if a row agrees with conditional statement.

    Return True if the row has to be skipped, otherwise it will return False.

    Parameters
    ----------
    row : dict
        Line parsed represented in a dict, where keys are fields.
    where : List[dict]
        A list of the conditional statements structured in a specified way.

    Returns
    -------
    bool
       Return True if the row has to be skipped and doesn't fulfill the conditional statement.
    """
    if where is None or len(where) == 0:
        return False

    filter_wh = False
    for k in where:
        try:
            value = row[k[WhereAttributesKeys.FIELD.value]]
            data_value = f"\"{value}\"" if isinstance(value, str) and not value.isnumeric() else str(value)
            filter_wh = eval(data_value + ' ' +
                             str(WHERE_STMTS_REVERSE[k[WhereAttributesKeys.OPERATION.value]]) + ' ' +
                             str(k[WhereAttributesKeys.VALUE.value]))
            return not filter_wh
        except (KeyError, ValueError):
            return True
    return filter_wh



"""
Variant
====================================
A core class to represent the files that will be parsed by an Annotation file.
"""

import csv
import ctypes
import gzip
import lzma
from fnmatch import fnmatch
from functools import lru_cache
import mmap
from os.path import isdir, isfile
import re
from typing import Generator, List, Callable, Any

from openvariant.annotation.annotation import Annotation
from openvariant.annotation.builder import MappingBuilder
from openvariant.annotation.process import AnnotationTypesProcess
from openvariant.config.config_annotation import AnnotationFormat, AnnotationTypes, AnnotationDelimiter


def _open_file(file_path: str, mode='r+b'):
    """Open raw files or compressed files"""

    if file_path.endswith('xz'):
        open_method = lzma.open
        file = open_method(file_path, mode)
        mm = file
    else:
        open_method = open
        file = open_method(file_path, mode)
        mm: mmap = mmap.mmap(file.fileno(), length=0, access=mmap.ACCESS_READ)

    return mm, file


def _base_parser(mm_obj: mmap, file_path: str, delimiter: str) -> Generator[int, str, None]:
    """Cleaning comments and irrelevant data"""
    try:
        if file_path.endswith('gz') or file_path.endswith('bgz'):
            mm_obj = gzip.GzipFile(mode="r+b", fileobj=mm_obj)
    except KeyError:
        raise KeyError(f"'{delimiter}' key not found.")

    for l_num, line in enumerate(iter(mm_obj.readline, b'')):
        line = line.decode('utf-8')
        row_line = line.split(AnnotationDelimiter[delimiter].value)

        if len(row_line) == 0:
            continue

        # Skip comments
        if (row_line[0].startswith('#') or row_line[0].startswith('##') or row_line[0].startswith('browser') or
            row_line[0].startswith('track')) and not row_line[0].startswith('#CHROM'):
            continue

        yield l_num, row_line


def _exclude(line: dict, excludes: dict) -> bool:
    """Excludes values described on the annotation file"""
    for k, v in excludes.items():
        for val in v:

            if val is not None and val.startswith("!"):
                valp = val.replace("!", "")
                if line[k] != valp:
                    return True
            else:
                if line[k] == val:
                    return True
    return False


def _extract_header(file_path: str, original_header: list, annotation: Annotation):
    """Extract header to parse the entire file, and create a reference for each field"""
    header_schema = {}
    mapping_fields = []

    for field, ann in annotation.annotations.items():
        ann_type = ann[0]
        if ann_type == AnnotationTypes.MAPPING.value:
            mapping_fields.append((field, ann))
        else:
            header_schema.update({field: AnnotationTypesProcess[ann_type].value(ann, original_header, file_path,
                                                                                header_schema)})

    for field, ann in mapping_fields:
        ann_type = ann[0]
        header_schema.update({field: AnnotationTypesProcess[ann_type].value(ann, original_header, file_path,
                                                                            header_schema)})
    return header_schema, annotation.columns


@lru_cache(maxsize=256)
def _parse_field(value: float or int or str, func: Callable) -> str:
    """Getting the value of a specific annotation field. Cached with LRU policy"""
    result = func(value)
    return result if result is not None else str(float('nan'))


def _parse_plugin_field(row: dict, field_name: str, file_path: str, value: Any, func: Callable) -> str:
    """Getting the value of a specific plugin annotation. No cached"""
    ctxt = value(row, field_name, file_path)
    return func(ctxt)


def _parse_mapping_field(x: MappingBuilder, row: dict, func: Callable):
    """Getting the value of a specific mapping annotation. No cached"""
    if x[1] is None:
        raise ValueError(f'Wrong source fields on {x[0]} annotation')
    value = None
    for source in x[1]:
        try:
            map_key = row[source]
            value = x[2].get(map_key, None)
        except KeyError:
            pass
    return str(value) if value is not None else str(float('nan'))


def _parser(file_path: str, annotation: Annotation, group_by: str, display_header: bool) \
        -> Generator[dict, None, None]:
    """Parsing of an entire file with annotation schema"""
    header, row, row_header = None, {}, []
    mm, file = _open_file(file_path, "rb")
    for lnum, line in _base_parser(mm, file_path, annotation.delimiter):
        if header is None:
            header, row_header = _extract_header(file_path, line, annotation)
            if not display_header:
                continue
            row = row_header
            yield row
        else:
            try:
                line_dict = {}
                row, plugin_values, mapping_values = {}, {}, {}
                for head in annotation.annotations.keys():
                    type_ann, value, func = header[head]
                    if type_ann == AnnotationTypes.PLUGIN.name:
                        plugin_values[head] = header[head]
                    elif type_ann == AnnotationTypes.MAPPING.name:
                        mapping_values[head] = header[head]
                    elif type_ann == AnnotationTypes.INTERNAL.name:
                        value = line[value] if value is not None else None
                        line_dict[head] = _parse_field(value, func)
                    else:
                        line_dict[head] = _parse_field(value, func)

                for head, mapping in mapping_values.items():
                    _, builder_mapping, func = mapping
                    line_dict[head] = _parse_mapping_field(builder_mapping, line_dict, func)
                for head, plug in plugin_values.items():
                    _, ctxt_plugin, func_plugin = plug
                    line_dict[head] = _parse_plugin_field(line_dict, head, file_path, ctxt_plugin, func_plugin)

                for k in annotation.columns:
                    row[k] = line_dict[k].format(**line_dict)

                if group_by is not None and group_by not in annotation.columns:
                    try:
                        row[group_by] = line_dict[group_by].format(**line_dict)
                    except KeyError as e:
                        raise KeyError(f"Unable to find group by: {e}. Check annotation for {file_path} file")

                if row and not _exclude(line_dict, annotation.excludes):
                    yield row

            except (ValueError, IndexError, KeyError) as e:
                mm.close()
                file.close()
                raise ValueError(f"Error parsing line: {lnum} {file_path}: {e}")
    mm.close()
    file.close()


def _check_extension(ext: str, path: str) -> bool:
    """Check if file matches with the annotation pattern"""
    if ext[0] == '*':
        match = fnmatch(path, ext)
    else:
        reg_apply = re.compile(ext + '$')
        match = len(reg_apply.findall(path)) != 0
    return match


def _unify(base_path: str, annotation: Annotation, group_by: str = None, display_header: bool = True) \
        -> Generator[dict, None, None]:
    """Parse all the files thought the annotation schema and generated yields to interrate"""
    for x in _parser(base_path, annotation, group_by, display_header):
        yield x


class Variant:
    """A representation of parsed files

        Methods
        -------
        read(group_key: str or None = None)
            Read the parsed files with its proper annotation.
        save(file_path: str, display_header: bool = True)
            Save parsed files on specified location.
    """

    def __init__(self, path: str, annotation: Annotation) -> None:
        """
        Inits Variant with files path and Annotation object

        Parameters
        ---------
        path : str
            A string path where files to parse are located (could be directory or a single file).
        annotation : Annotation
            Object to describe the schema of parsed files.
        """
        if path is None or path == '' or not isfile(path):
            raise ValueError('Invalid path, must be a file')
        if annotation is None:
            raise ValueError('Invalid annotation')

        csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))
        self._path: str = path
        self._annotation: Annotation = annotation
        self._header: List[str] = list(annotation.annotations.keys()) if len(annotation.columns) == 0 \
            else annotation.columns

    @property
    def path(self) -> str:
        """str: Path where parsed files are located"""
        return self._path

    @property
    def header(self) -> List[str]:
        """List[str]: Header of the corresponding parsed files"""
        return self._header

    @property
    def annotation(self) -> Annotation:
        """Annotation: Annotation object which files were parsed"""
        return self._annotation

    def read(self, group_key: str or None = None) -> Generator[dict, None, None]:
        """
        Read parsed files and generated an iterator for each row

        Parameters
        ---------
        group_key : str or None
            A string that indicates how rows will be grouped (optional).

        Yields
        ------
        dict
            Representation of a parsed row.
        """
        for i, line in enumerate(_unify(self._path, self._annotation, group_by=group_key)):
            if i != 0:
                yield line

    def save(self, file_path: str, display_header: bool = True) -> None:
        """
        Save parsed files in an indicated location.

        Parameters
        ---------
        file_path : str or None
            A string that indicates the location to store the output file.

        display_header : bool
            A bool that indicates if the output will have header or not (optional).
        """
        if file_path is None or isdir(file_path):
            raise ValueError("The path must be a file.")
        with open(file_path, "w") as file:
            writer = csv.writer(file, delimiter=AnnotationFormat[self._annotation.format.upper()].value)
            for i, line in enumerate(_unify(self._path, self._annotation)):
                if display_header and i == 0:
                    writer.writerow(line)
                elif i != 0:
                    writer.writerow(line.values())



"""
Plugin task
====================================
A core functionality to execute different plugin tasks.
"""
import os
from enum import Enum
from functools import partial


def _add_action(name: str, directory: str = None) -> None:
    """Create a new plugin

    It will generate all the required structure for a new plugin (files and folders).

    Parameters
    ----------
    name : str
        Name of the plugin to create.
    directory : str
        The path to create the plugin.
    """
    path = directory if directory is not None else os.getcwd()
    if os.path.exists(f"{path}/{name}"):
        raise FileExistsError(f"Directory {path}/{name} already exists.")
    os.mkdir(f"{path}/{name}")
    plugin_name = f"{name.capitalize()}Plugin"
    context_name = f"{name.capitalize()}Context"
    with open(f"{path}/{name}/__init__.py", 'w') as init_file:
        init_file.write(f"import .{name} from {plugin_name}\n")
        init_file.write(f"import .{name} from {context_name}\n")
        init_file.close()
    with open(f"{path}/{name}/{name}.py", 'w') as plugin_file:
        plugin_file.write(f"from openvariant.plugins.context import Context\n")
        plugin_file.write(f"from openvariant.plugins.plugin import Plugin\n\n\n")
        plugin_file.write(f"class {context_name}(Context):\n")
        plugin_file.write(f"\t\"\"\"\n")
        plugin_file.write(f"\t{plugin_name} class generated by OpenVariant\n\n")
        plugin_file.write(f"\tAttributes\n")
        plugin_file.write(f"\t-------\n")
        plugin_file.write(f"\trow : dict\n")
        plugin_file.write(f"\t\tMain method to execute data transformation in each row.\n")
        plugin_file.write(f"\tfield_name : str\n")
        plugin_file.write(f"\t\tName of the corresponding column that was described on the annotation schema.\n")
        plugin_file.write(f"\tfile_path : str\n")
        plugin_file.write(f"\t\tPath of the Input file that is being parsed.\n")
        plugin_file.write(f"\t\"\"\"\n\n")
        plugin_file.write(f"\tdef __init__(self, row: dict, field_name: str, file_path: str) -> None:\n")
        plugin_file.write(f"\t\tsuper().__init__(row, field_name, file_path)\n")
        plugin_file.write(f"\n\n")
        plugin_file.write(f"class {plugin_name}(Plugin):\n")
        plugin_file.write(f"\t\"\"\"\n")
        plugin_file.write(f"\t{plugin_name} class generated by OpenVariant\n\n")
        plugin_file.write(f"\tMethods\n")
        plugin_file.write(f"\t-------\n")
        plugin_file.write(f"\trun(context: {context_name})\n")
        plugin_file.write(f"\t\tMain method to execute data transformation in each row.\n")
        plugin_file.write(f"\t\"\"\"\n\n")
        plugin_file.write(f"\tdef run(self, context: {context_name}) -> dict:\n")
        plugin_file.write(f"\t\t\"\"\"\n")
        plugin_file.write(f"\t\tData transformation of a single row\n\n")
        plugin_file.write(f"\t\tParameters\n")
        plugin_file.write(f"\t\t-------\n")
        plugin_file.write(f"\t\tcontext : {context_name}\n")
        plugin_file.write(f"\t\t\tRepresentation of the row to be parsed.\n")
        plugin_file.write(f"\t\t\n")
        plugin_file.write(f"\t\tReturns\n")
        plugin_file.write(f"\t\t-------\n")
        plugin_file.write(f"\t\tfloat or int or str\n")
        plugin_file.write(f"\t\t\tThe value of the field transformed.\n")
        plugin_file.write(f"\t\t\"\"\"\n")
        plugin_file.write(f"\n")
        plugin_file.write(f"\t\treturn context.row[context.field_name]\n")


class PluginActions(Enum):
    """Enum to get the function of each action"""

    """Create a new plugin"""
    CREATE = partial(_add_action)



class Context(object):
    """Base class that each context must inherit from"""

    def __init__(self, row: dict, field_name: str, file_path: str) -> None:
        self._row = row
        self._field_name = field_name
        self._file_path = file_path

    @property
    def row(self) -> dict:
        return self._row

    @property
    def field_name(self) -> str:
        return self._field_name

    @property
    def file_path(self) -> str:
        return self._file_path



import time
from os import cpu_count
import click as click

import openvariant
from openvariant.commands.tasks.cat import cat as cat_task
from openvariant.commands.tasks.count import count as count_task
from openvariant.commands.tasks.groupby import group_by as group_by_task
from openvariant.commands.tasks.plugin import PluginActions


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(openvariant.__version__)
def openvar():
    """'openvar' is the command-line interface of OpenVariant.
    Parsing and data transformation of multiple input formats."""
    pass


@openvar.command(name="cat", short_help='Concatenate files to standard input')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--annotations', '-a', type=click.Path(exists=True), default=None)
@click.option('--header', help="Show the result header", is_flag=True)
def cat(input_path: str, where: str or None, annotations: str or None, header: bool):
    """Print the parsed files on the stdout."""
    cat_task(input_path, annotations, where, header)


@openvar.command(name="count", short_help='Number of rows that matches a specified criterion')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', multiple=False, type=click.STRING, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--group_by', '-g', type=click.STRING, help="Filter expression. eg: CHROMOSOME")
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True))
@click.option('--cores', '-c', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def count(input_path: str, where: str, group_by: str, cores: int, quite: bool, annotations: str or None) -> None:
    """Print on the stdout the number of rows that meets the criteria."""
    result = count_task(input_path, annotations, group_by=group_by, where=where, cores=cores, quite=quite)
    if len(result[1]) > 0:
        for k, v in sorted(result[1].items(), key=lambda res: res[1]):
            print("{}\t{}".format(k, v))

    print("TOTAL\t{}".format(result[0]))


@openvar.command(name="groupby", short_help='Groups rows that have the same values into summary rows')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--header', help='Send header as first row', is_flag=True)
@click.option('--show', help='Show group by each row', is_flag=True)
@click.option('--group_by', '-g', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME")
@click.option('--where', '-w', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--script', '-s', type=click.STRING, default=None,
              help="Filter expression. eg: gzip > \${GROUP_KEY}.parsed.tsv.gz")
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True))
@click.option('--cores', '-c', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def groupby(input_path: str, script: str, where: str, group_by: str, cores: int, quite: bool, annotations: str or None,
            header: bool, show: bool):
    """Print on the stdout the parsed files group by a specified field."""
    for group_key, group_result, command in group_by_task(input_path, annotations, script, key_by=group_by, where=where,
                                                          cores=cores, quite=quite, header=header):
        for r in group_result:
            if command:
                print(f"{group_key}\t{r}") if show else print(f"{r}")
            else:
                if header:
                    print(f"{r}")
                    header = False
                else:
                    print(f"{group_key}\t{r}") if show else print(f"{r}")


@openvar.command(name="plugin", short_help='Actions to execute for a plugin: create')
@click.argument('action', type=click.Choice(['create']))
@click.option('--name', '-n', type=click.STRING)
@click.option('--directory', '-d', type=click.STRING)
def plugin(action, name: str or None, directory: str or None):
    """Actions to apply on the plugin system."""
    PluginActions[action.upper()].value(name, directory)


if __name__ == "__main__":
    openvar()



from typing import Tuple

from openvariant.plugins.context import Context
from openvariant.plugins.plugin import Plugin


def _prefix_length(ref: str, alt: str) -> int:
    i = 0
    while i < len(ref) and i < len(alt) and ref[i] == alt[i]:
        i += 1
    return i


def _suffix_length(ref: str, alt: str) -> int:
    i = len(ref) - 1
    j = len(alt) - 1
    while i >= 0 and j >= 0 and ref[i] == alt[j]:
        i -= 1
        j -= 1
    return len(ref) - i - 1


def _indel_postprocess(start: int, ref: str, alt: str) -> Tuple[str, str, str]:
    """Removes the bases that are repeated in both ref and alt and are therefore NOT variants"""
    prefix_len = _prefix_length(ref, alt)
    ins_correction = 1 if len(ref) < len(alt) else 0
    start = int(start) + max(0, prefix_len - ins_correction)
    alt = alt[prefix_len:]
    ref = ref[prefix_len:]

    suffix_len = _suffix_length(ref, alt)
    if suffix_len > 0:
        alt = alt[:-suffix_len]
        ref = ref[:-suffix_len]

    ref = '-' if ref == '' else ref
    alt = '-' if alt == '' else alt

    return str(start), str(ref), str(alt)


class Alteration_typeContext(Context):
    """
    The context of this plugin will be the simplest one, without any added property or methods.

    Attributes
    ----------
    row : dict
        The row that data transformation will be applied.
    field_name : str
        Name of the corresponding column that was described on the annotation schema.
    file_path : str
        Path of the Input file that is being parsed.
    """

    def __init__(self, row: dict, field_name: str, file_path: str) -> None:
        super().__init__(row, field_name, file_path)


class Alteration_typePlugin(Plugin):
    """
    This plugin identifies the alteration type.
    Classifies the alteration type: checking POSITION, REF and ALT fields.
    The result will be store in a field of the input.
    """

    def run(self, context: Alteration_typeContext) -> str:
        """Extract alteration type from the input row.

        Parameters
        ----------
        context : Alteration_typeContext
            Representation of the row to be parsed.

        Returns
        -------
        str
            The value of ALT_TYPE field.
        """
        row = context.row
        if 'REF' in row and 'ALT' in row:
            l_ref = len(row['REF'])
            l_alt = len(row['ALT'])

            if l_alt != l_ref:
                alt_type = "indel"
            else:
                if l_alt > 1:
                    alt_type = "mnv"
                else:
                    if '-' in row['REF'] or '-' in row['ALT']:
                        alt_type = "indel"
                    else:
                        alt_type = "snv"
            if alt_type == "indel":
                row['POSITION'], row['REF'], row['ALT'] = _indel_postprocess(row['POSITION'], row['REF'], row['ALT'])

            row[context.field_name] = alt_type
        else:
            raise ValueError("Unable to find 'REF', 'ALT' or 'POSITION' values in the row.")
        return row[context.field_name]



from openvariant.plugins.plugin import Plugin


class BasicPlugin(Plugin):
    """
    Basic template of a customized plugin.
    This is a subclass of Plugin class, implemented inside OpenVariant.
    Use this files as a guidance to build new ones.
    """

    def run(self, row: dict) -> dict:
        return row



class Plugin(object):
    """Base class that each plugin must inherit from"""

    def run(self, argument):
        """
        This is a main function (required) which OpenVariant will call every time that this plugin is
        described in an annotation file.
        """
        raise NotImplementedError



"""
Cat  task
====================================
A core functionality to execute cat task.
"""
from typing import List

from openvariant.config.config_annotation import AnnotationFormat
from openvariant.find.find import find_files
from openvariant.utils.where import parse_where, skip
from openvariant.variant.variant import Variant


def _format_line(line: List[str], out_format: str) -> str:
    """Line formatting for output"""
    return AnnotationFormat[out_format.upper()].value.join(line)


def cat(base_path: str, annotation_path: str or None = None, where: str = None, header_show: bool = True) -> None:
    """Print on the stdout the parsed files.

    It will parse the input files with its proper annotation schema, and it'll show the result on the stdout.
    It can be printed with or without header. Can be added a 'where' expression.

    Parameters
    ----------
    base_path : srt
        Base path of input files.
    annotation_path : str or None
        Path of annotation file.
    where : str
        Conditional statement.
    header_show : bool
        Shows header on the output.
    """
    for file, annotation in find_files(base_path, annotation_path):
        where_clauses = parse_where(where)
        result = Variant(file, annotation)
        header = result.header
        if header_show:
            print(_format_line(header, result.annotation.format))
            header_show = False
        for i, r in enumerate(result.read()):
            if isinstance(r, dict):
                if skip(r, where_clauses):
                    continue
                print(_format_line(list(map(str, r.values())), result.annotation.format))



import re
from enum import Enum
from functools import partial
from os.path import isdir, basename, abspath, dirname
from typing import Tuple, List, Callable, Optional

from openvariant.annotation.builder import StaticBuilder, InternalBuilder, FilenameBuilder, Builder, DirnameBuilder, \
    PluginBuilder, MappingBuilder
from openvariant.config.config_annotation import AnnotationTypes
from openvariant.plugins.context import Context

StaticProcess = Tuple[str, float or int or str, Callable]
InternalProcess = Tuple[str, Optional[int], Callable]
FilenameProcess = Tuple[str, float or int or str, Callable]
DirnameProcess = Tuple[str, float or int or str, Callable]
PluginProcess = Tuple[str, Context, Callable]
MappingProcess = Tuple[str, MappingBuilder, Callable]


def _static_process(x: StaticBuilder, original_header: List = [] or None, file_path: str = None,
                    annotation: dict = None) \
        -> StaticProcess:
    """Get a Static value
    It will return a StaticProcess describing the value to get from static annotation.
    Parameters
    ----------
    x : StaticBuilder
        Annotation builder.
    Returns
    -------
    str
        Annotation type
    float or int or str
        Fixed value
    Callable
        Function to execute on the fixed value
    """
    try:
        return AnnotationTypes.STATIC.name, x[1] if x[1] is not None else float('nan'), str
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')


def _internal_process(x: InternalBuilder, original_header: List = [] or None, file_path: str = None,
                      annotation: dict = None) \
        -> InternalProcess:
    """Get an Internal value
    It will return a InternalProcess describing the value to get from internal annotation.
    Parameters
    ----------
    x : InternalBuilder
        Annotation builder.
    Returns
    -------
    str
        Annotation type
    float or int or str
        Fixed value
    Callable
        Function to execute on the fixed value
    """
    field_pos = None
    try:
        for i, h in enumerate(original_header):
            if h in set(x[1]):
                field_pos = i
                break
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')

    return AnnotationTypes.INTERNAL.name, field_pos, x[2]


def _filename_process(x: FilenameBuilder, original_header: List = [] or None, file_path: str = None,
                      annotation: dict = None) \
        -> FilenameProcess:
    """Get a Filename value
    It will return a FilenameProcess describing the value to get from filename annotation.
    Parameters
    ----------
    x : FilenameBuilder
        Annotation builder
    file_path: str
        Path of input file
    Returns
    -------
    str
        Annotation type
    float or int or str
        Fixed value
    Callable
        Function to execute on the fixed value
    """
    try:
        if isdir(file_path):
            raise FileNotFoundError('Unable to find a filename')

        func_result = x[1](basename(file_path))
        value = x[2].findall(func_result)[0]
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')
    except (re.error, IndexError):
        raise re.error(f'Wrong regex pattern on {x[0]} annotation')

    return AnnotationTypes.FILENAME.name, value if value is not None else float('nan'), str


def _dirname_process(x: DirnameBuilder, original_header: List = [] or None, file_path: str = None,
                     annotation: dict = None) \
        -> DirnameProcess:
    """Get a Dirname value
    It will return a DirnameProcess describing the value to get from dirname annotation.
    Parameters
    ----------
    x : DirnameBuilder
        Annotation builder
    file_path: str
        Path of input file
    Returns
    -------
    str
        Annotation type
    float or int or str
        Fixed value
    Callable
        Function to execute on the fixed value
    """
    try:
        if isdir(file_path):
            raise FileNotFoundError('Unable to find a dirname')

        func_result = x[1](basename(dirname(abspath(file_path))))
        value = x[2].findall(func_result)[0]
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')
    except (re.error, IndexError):
        raise re.error(f'Wrong regex pattern on {x[0]} annotation')

    return AnnotationTypes.DIRNAME.name, value if value is not None else float('nan'), str


def _mapping_process(x: MappingBuilder, original_header: List = [] or None, file_path: str = None,
                     header_schema: dict = None) \
        -> MappingProcess:
    """Get a Mapping value
    It will return a PluginProcess describing the value to get from plugin annotation.
    Parameters
    ----------
    x : PluginBuilder
        Annotation builder
    Returns
    -------
    str
        Annotation type
    None
        A None value, this value will not be taken into account
    Callable
        Function to execute on the fixed value
    """

    """
    if x[1] is None:
        raise ValueError(f'Wrong source fields on {x[0]} annotation')
    value = None

    for source in x[1]:
        try:
            map_key = header_schema[source][1]
            value = x[2].get(map_key, None)
        except KeyError:
            pass

    if value is None:
        raise KeyError(f'Unable to map {x[1]} sources on mapping annotation')
    """
    return AnnotationTypes.MAPPING.name, x, str #value if value is not None else float('nan'), str


def _plugin_process(x: PluginBuilder, original_header: List = [] or None, file_path: str = None,
                    annotation: dict = None) \
        -> PluginProcess:
    """Get a Plugin value
    It will return a PluginProcess describing the value to get from plugin annotation.
    Parameters
    ----------
    x : PluginBuilder
        Annotation builder
    Returns
    -------
    str
        Annotation type
    None
        A None value, this value will not be taken into account
    Callable
        Function to execute on the fixed value
    """
    if x[1] is None or x[2] is None:
        raise ValueError(f'Wrong function on {x[0]} annotation')
    return AnnotationTypes.PLUGIN.name, x[2], x[1]


class AnnotationTypesProcess(Enum):
    """Enum to get the value of every annotation Builder"""

    """Parser for static builder"""
    STATIC = partial(_static_process)

    """Parser for internal builder"""
    INTERNAL = partial(_internal_process)

    """Parser for filename builder"""
    FILENAME = partial(_filename_process)

    """Parser for dirname builder"""
    DIRNAME = partial(_dirname_process)

    """Parser for mapping builder"""
    MAPPING = partial(_mapping_process)

    """Parser for plugin builder"""
    PLUGIN = partial(_plugin_process)



"""
Find
====================================
A core functionally to search files and with its corresponding annotation.
"""
import glob
import re
from fnmatch import fnmatch
from os import listdir
from os.path import isfile, join, basename, abspath, dirname
from typing import Generator
from copy import deepcopy

from openvariant.annotation.annotation import Annotation

from openvariant.config.config_annotation import ANNOTATION_EXTENSION


def _check_extension(ext: str, path: str) -> bool:
    """Check if file matches with the annotation pattern"""
    if ext[0] == '*':
        match = fnmatch(path, ext)
    else:
        reg_apply = re.compile(ext + '$')
        match = len(reg_apply.findall(path)) != 0
    return match


def _find_files(base_path: str, annotation: Annotation or None, fix: bool) -> Generator[str, Annotation, None]:
    """Recursive exploration from a base path"""
    if not fix:
        for annotation_file in glob.iglob(join(base_path, "*.{}".format(ANNOTATION_EXTENSION))):
            annotation = Annotation(annotation_file)

    for file_name in listdir(base_path):
        try:
            file_path = join(base_path, file_name)

            if isfile(file_path):

                for ext, _ in annotation.structure.items():
                    if _check_extension(ext, file_path):
                        yield file_path, annotation
            else:
                for f, a in _find_files(file_path, annotation, fix):
                    yield f, a

        except PermissionError as e:
            raise PermissionError(f"Unable to open {file_name}: {e}")


def find_files(base_path: str, annotation_path: str or None = None) -> Generator[str, Annotation, None]:
    """Get each file and its proper annotation object.

    Parameters
    ----------
    base_path : srt
        Base path of input files.
    annotation_path : str or None
        Path of annotation file.

    Yields
    -------
    str
        Input file's name.
    Annotation
        The proper schema of each input file.
    """
    annotation, fix = (Annotation(annotation_path), True) if annotation_path is not None else (None, False)
    return _find_files(base_path, annotation, fix)



"""
Annotation
====================================
A core class to represent the schema which files will be parsed.
"""
import logging
import re

from typing import List
from yaml import safe_load, YAMLError

from openvariant.annotation.builder import AnnotationTypesBuilders
from openvariant.config.config_annotation import (AnnotationGeneralKeys, AnnotationKeys, AnnotationTypes,
                                                  ExcludesKeys, DEFAULT_FORMAT, DEFAULT_DELIMITER,
                                                  AnnotationFormat, AnnotationDelimiter)


def _check_general_keys(annot: dict) -> None:
    """Check if general annotations are writen in a proper format"""
    # Pattern key
    if AnnotationGeneralKeys.PATTERN.value not in annot or not isinstance(
            annot[AnnotationGeneralKeys.PATTERN.value], list) \
            and not all(isinstance(x, str) for x in annot[AnnotationGeneralKeys.PATTERN.value]):
        raise KeyError(f"'{AnnotationGeneralKeys.PATTERN.value}' key not found or is not a str.")

    # Recursive key
    if AnnotationGeneralKeys.RECURSIVE.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.RECURSIVE.value], bool):
        raise KeyError(f"'{AnnotationGeneralKeys.RECURSIVE.value}' key is not a boolean.")

    # Format key
    if AnnotationGeneralKeys.FORMAT.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.FORMAT.value], str) or
             annot[AnnotationGeneralKeys.FORMAT.value].upper() not in [e.name for e in AnnotationFormat]):
        raise KeyError(f"'{AnnotationGeneralKeys.FORMAT.value}' key is not a string.")

    # Delimiter key
    if AnnotationGeneralKeys.DELIMITER.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.DELIMITER.value], str) or
             annot[AnnotationGeneralKeys.DELIMITER.value].upper() not in [e.name for e in AnnotationDelimiter]):
        raise KeyError(f"'{AnnotationGeneralKeys.DELIMITER.value}' key is not valid or is not a string.")

    # Columns key
    if AnnotationGeneralKeys.COLUMNS.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.COLUMNS.value], list):
        raise KeyError(f"'{AnnotationGeneralKeys.COLUMNS.value}' key is not a list.")

    # Annotations key
    if AnnotationGeneralKeys.ANNOTATION.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.ANNOTATION.value], list):
        raise KeyError(f"'{AnnotationGeneralKeys.ANNOTATION.value}' key is not a list.")

    # Excludes key
    if AnnotationGeneralKeys.EXCLUDE.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.EXCLUDE.value], list) or
             not all([ExcludesKeys.FIELD.value in x and ExcludesKeys.VALUE.value in x
                      for x in annot[AnnotationGeneralKeys.EXCLUDE.value]])):
        raise KeyError(f"'{AnnotationGeneralKeys.EXCLUDE.value}' key in bad format.")


def _check_annotation_keys(annot: dict) -> None:
    """Check if annotation keys are writen in a proper format"""
    # Type key
    if AnnotationKeys.TYPE.value not in annot or not isinstance(annot[AnnotationKeys.TYPE.value], str):
        raise KeyError(f"'{AnnotationKeys.TYPE.value}' key not found or is not a str.")
    if annot[AnnotationKeys.TYPE.value] not in [e.value for e in AnnotationTypes]:
        raise ValueError(f"'{AnnotationKeys.TYPE.value}' value is wrong.")

    # Field key
    if AnnotationKeys.FIELD.value not in annot or not isinstance(annot[AnnotationKeys.FIELD.value], str):
        raise KeyError(f"'{AnnotationKeys.FIELD.value}' key not found or is not a str.")

    # Value key
    if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.STATIC.value) and \
            not isinstance(annot[AnnotationKeys.VALUE.value], str):
        raise KeyError(f"'{AnnotationKeys.VALUE.value}' key not found or is not a str.")

    # Field source key
    if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.INTERNAL.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.MAPPING.value) and \
            AnnotationKeys.FIELD_SOURCE.value in annot and \
            not isinstance(annot[AnnotationKeys.FIELD_SOURCE.value], list):
        raise KeyError(f"'{AnnotationKeys.FIELD_SOURCE.value}' key not found or is not a list.")

    # Dirname and filename key
    if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.DIRNAME.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.FILENAME.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.INTERNAL.value) and \
            AnnotationKeys.FUNCTION.value in annot and \
            re.compile("lambda[' ']+[a-zA-Z0-9]+[' ']*:[' ']*.*").search(annot[AnnotationKeys.FUNCTION.value]) is None:
        raise ValueError(f"'{AnnotationKeys.FUNCTION.value}' value is not an appropriated lambda function.")

    # Plugin key
    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value and \
            AnnotationKeys.PLUGIN.value not in annot:
        raise KeyError(f"'{AnnotationKeys.PLUGIN.value}' key not found.")

    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value and \
            (AnnotationKeys.PLUGIN.value in annot and not isinstance(annot[AnnotationKeys.PLUGIN.value], str)):
        raise ValueError(f"'{AnnotationKeys.PLUGIN.value}' is not a str.")

    # Mapping keys
    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.MAPPING.value and \
            (AnnotationKeys.FIELD_SOURCE.value not in annot or
             AnnotationKeys.FIELD_MAPPING.value not in annot or
             AnnotationKeys.FILE_MAPPING.value not in annot or
             AnnotationKeys.FIELD_VALUE.value not in annot or
             not isinstance(annot[AnnotationKeys.FIELD_SOURCE.value], list) or
             not isinstance(annot[AnnotationKeys.FIELD_MAPPING.value], str) or
             not isinstance(annot[AnnotationKeys.FILE_MAPPING.value], str) or
             not isinstance(annot[AnnotationKeys.FIELD_VALUE.value], str)):
        raise KeyError(f"'{AnnotationTypes.MAPPING.value}' not annotated well.")


class Annotation:
    """A representation of the schema that files will be parsed"""

    def _read_annotation_file(self) -> dict:
        """Read annotation file with YAML package"""
        with open(self._path, 'r') as stream:
            try:
                return safe_load(stream)
            except YAMLError as exc:
                logging.error(exc)
            stream.close()

    def __init__(self, annotation_path: str) -> None:
        """
        Inits Annotation with annotation file path.

        Parameters
        ---------
        annotation_path : str
            A string path where Annotation file is located.
        """

        self._path = annotation_path
        raw_annotation = self._read_annotation_file()
        _check_general_keys(raw_annotation)
        for annot in raw_annotation.get(AnnotationGeneralKeys.ANNOTATION.value, []):
            _check_annotation_keys(annot)

        patterns = raw_annotation[AnnotationGeneralKeys.PATTERN.value]
        self._patterns = patterns if isinstance(patterns, List) else [patterns]
        self._recursive = raw_annotation.get(AnnotationGeneralKeys.RECURSIVE.value, True)
        self._delimiter = raw_annotation.get(AnnotationGeneralKeys.DELIMITER.value, DEFAULT_DELIMITER).upper()

        self._format = raw_annotation.get(AnnotationGeneralKeys.FORMAT.value, DEFAULT_FORMAT).replace('.', '')

        self._excludes: dict = {}

        for k in raw_annotation.get(AnnotationGeneralKeys.EXCLUDE.value, []):
            key_exclude = k[AnnotationKeys.FIELD.value]
            value_exclude = k[AnnotationKeys.VALUE.value]
            if key_exclude in self._excludes:
                self._excludes[key_exclude].append(value_exclude)
            else:
                self._excludes[key_exclude] = [value_exclude]

        self._annotations: dict = {}
        for k in raw_annotation.get(AnnotationGeneralKeys.ANNOTATION.value, []):

            self._annotations[k[AnnotationKeys.FIELD.value]] = \
                AnnotationTypesBuilders[k[AnnotationKeys.TYPE.value].upper()].value(k, self._path)

        self._columns = raw_annotation.get(AnnotationGeneralKeys.COLUMNS.value, list(self.annotations.keys()))
        self._check_columns()

    def _check_columns(self) -> None:
        """Check if columns exists as annotation fields"""
        for col in self._columns:
            if col not in self._annotations:
                raise KeyError(f"'{col}' column unable to find.")

    @property
    def path(self) -> str:
        """str: path where annotation file is located"""
        return self._path

    @property
    def patterns(self) -> List[str]:
        """List[str]: files patterns that annotation will match"""
        return self._patterns

    @property
    def format(self) -> str:
        """str: output format that will have parsed files"""
        return self._format

    @property
    def delimiter(self) -> str:
        """str: delimiter that annotation will read on files"""
        return self._delimiter

    @property
    def columns(self) -> List:
        """List: columns that will appear on parsed output files"""
        return self._columns

    @property
    def annotations(self) -> dict:
        """dict: annotation that will cover Annotation object"""
        return self._annotations

    @property
    def excludes(self) -> dict:
        """List: values that will be excluded after the parsing"""
        return self._excludes

    @property
    def structure(self) -> dict:
        """dict: general structure of Annotation schema"""
        structure_aux = {AnnotationGeneralKeys.ANNOTATION.name: self._annotations,
                         AnnotationGeneralKeys.EXCLUDE.name: self._excludes}
        return {e: structure_aux for e in self._patterns}



"""
Group by task
====================================
A core functionality to execute group by task.
"""
from collections import defaultdict
from functools import partial
from multiprocessing import Pool
from os import cpu_count
from subprocess import PIPE, Popen
from typing import Generator, List, Tuple

from tqdm import tqdm

from openvariant.annotation.annotation import Annotation
from openvariant.find.find import find_files
from openvariant.utils.logger import log
from openvariant.utils.where import skip, parse_where
from openvariant.variant.variant import Variant


def _get_unique_values(file_path: str, annotation: Annotation, key: str) -> tuple[set, list]:
    """Get unique values of the group by field"""
    values = set()
    result = Variant(file_path, annotation)
    result_read = []
    try:
        for r in result.read(key):
            values.add(r[key])
            result_read.append(r)
    except KeyError:
        log.warn(f"'{key}' key not found in '{file_path}' file")
    return values, result_read


def _group(base_path: str, annotation_path: str or None, key_by: str) -> list[tuple[str, List]]:
    """Group file and its annotation by the group value"""
    results = defaultdict(list)
    for file, ann in find_files(base_path, annotation_path):
        # print(file, ann)
        by_value = ann.annotations.get(key_by, None)
        # print(by_value)

        if isinstance(by_value, tuple):
            values, result_read = _get_unique_values(file, ann, key_by)
            for s in values:
                results[s].append((file, ann))

    results_by_groups = []
    for key, group_select in results.items():
        results_by_groups.append((key, group_select))
    return results_by_groups

def _group_by_task(selection, where=None, key_by=None, script='', header=False) -> Tuple[str, List, bool]:
    """Main functionality for group by task"""
    where_clauses = parse_where(where)
    group_key, group_values = selection

    output = []
    if script is None:
        try:
            for value in group_values:
                result = Variant(value[0], value[1])

                columns = result.annotation.columns if len(result.annotation.columns) != 0 else result.header
                # columns = []
                if header:
                    line = "\t".join([str(h).strip() for h in columns])
                    output.append(f"{line}")
                    header = False

                # for row in value[2]:
                #    print(row)

                for row in result.read(key_by):

                    if skip(row, where_clauses):
                        continue

                    if row[key_by] == group_key:
                        line = "\t".join([str(row[h]).strip() for h in columns])
                        output.append(f"{line}")
                    # break
        except BrokenPipeError:
            pass
        return group_key, output, False

    else:
        try:
            process = Popen(script, shell=True, stdin=PIPE, stdout=PIPE,
                            env={"GROUP_KEY": 'None' if group_key is None else group_key})
        except ProcessLookupError as e:
            raise ChildProcessError(f"Unable to run '{script}': {e}")
        try:
            for value in group_values:
                result = Variant(value[0], value[1])
                columns = result.annotation.columns if len(result.annotation.columns) != 0 else result.header

                if header:
                    line = "\t".join([str(h).strip() for h in columns])
                    process.stdin.write(f"{line}\n".encode())
                    process.stdin.flush()
                    header = False
                for row in result.read(key_by):
                    if skip(row, where_clauses):
                        continue
                    try:
                        if row[key_by] == group_key:
                            line = "\t".join([str(row[h]).strip() for h in columns])
                            process.stdin.write(f"{line}\n".encode())
                            process.stdin.flush()
                    except KeyError:
                        pass
            process.stdin.close()
        except BrokenPipeError:
            pass

        try:
            while True:
                out = process.stdout.readline().decode().strip()
                if out == "":
                    break
                output.append(out)
            process.stdout.close()
        except BrokenPipeError:
            pass
        return group_key, output, True


def group_by(base_path: str, annotation_path: str or None, script: str or None, key_by: str, where: str or None = None,
             cores=cpu_count(), quite=False, header: bool = False) -> Generator[Tuple[str, List, bool], None, None]:
    """Print on the stdout the group by result.

    It'll parse the input files with its proper annotation schema, and it'll show the parsed result separated for each
    group by value. It'll be grouped by a field and can be added a 'where' expression. Also, the result can be
    executed thought a bash script.

    Parameters
    ----------
    base_path : srt
        Base path of input files.
    annotation_path : str or None
        Path of annotation file.
    script : str or None
        Path of annotation file.
    key_by : str
        Field to group the result.
    where : str
        Conditional statement.
    quite : bool
        Discard progress bar.
    cores : int
        Number of cores to parallelize the task.
    header : bool
        Number of cores to parallelize the task.
    Returns
    ----------
    int
        The total number of rows.
    dict
        A schema with separate groups and the numbers of rows for each.
    """
    selection = _group(base_path, annotation_path, key_by)
    with Pool(cores) as pool:
        task = partial(_group_by_task, where=where, key_by=key_by, script=script, header=header)
        map_method = map if cores == 1 or len(selection) <= 1 else pool.imap_unordered

        for group_key, group_result, command in tqdm(
                map_method(task, selection),
                total=len(selection),
                desc="Computing groups".rjust(40),
                disable=(len(selection) < 2 or quite)
        ):
            yield group_key, group_result, command



"""
Builder
====================================
A core Enum to build a specified Tuple for each annotation.
"""
import csv
import inspect
import os
import re
import glob
import gzip
import importlib
import importlib.util
from enum import Enum
from functools import partial
from os.path import dirname
from typing import Tuple, Any, List, Callable

from openvariant.config.config_annotation import AnnotationKeys, AnnotationTypes
from openvariant.plugins.context import Context
from openvariant.plugins.plugin import Plugin


class Builder:
    """A representation of a function in annotation file"""
    func: str = None

    def __init__(self, func: str) -> None:
        """
        Inits Builder with function or lambda in str format.

        Parameters
        ---------
        func : str
            A string that represents a function or a lambda described on the annotation file.
        """
        self.func = func

    def __call__(self, x: Any) -> Any:
        return eval(self.func)(x)


StaticBuilder = Tuple[str, float or int or str]
InternalBuilder = Tuple[str, List, Builder, str or float]
DirnameBuilder = Tuple[str, Builder, re.Pattern]
FilenameBuilder = Tuple[str, Builder, re.Pattern]
PluginBuilder = Tuple[str, Callable, Context]
MappingBuilder = Tuple[str, List, dict]


def _get_function_and_regexp(x: dict) -> Tuple[Builder, re.Pattern]:
    """Get the function and regular expression of an annotation
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    Builder
        Represents the function described on the annotation.
    re.Pattern
        Regular expression to parse the annotation.
    """
    func_apply = Builder("(lambda y: y)") if AnnotationKeys.FUNCTION.value not in x \
        else Builder(x[AnnotationKeys.FUNCTION.value])
    try:
        regex_apply = re.compile('(.*)') if AnnotationKeys.REGEX.value not in x or x[AnnotationKeys.REGEX.value] is None \
            else re.compile(x[AnnotationKeys.REGEX.value])
    except re.error as e:
        raise re.error(f'Wrong regex pattern: {e}')
    return func_apply, regex_apply


def _static_builder(x: dict, base_path: str = None) -> StaticBuilder:
    """Built StaticBuilder from an annotation based on a static annotation with a fixed value.
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    str
        Annotation type
    float or int or str
        Value of the field
    """
    try:
        value = x[AnnotationKeys.VALUE.value]
    except KeyError:
        raise KeyError('Static annotation is wrong specified.')
    return AnnotationTypes.STATIC.name, value


def _internal_builder(x: dict, base_path: str = None) -> InternalBuilder:
    """Built InternalBuilder from an annotation based on an internal annotation from fields of input files.
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    str
        Annotation type
    str
        Value of the field
    Builder
        Representation of the function to apply on the annotation value
    """
    try:
        value = x[AnnotationKeys.VALUE.value]
    except KeyError:
        value = float('nan')

    return AnnotationTypes.INTERNAL.name, x[AnnotationKeys.FIELD_SOURCE.value], Builder("(lambda y: y)") \
        if AnnotationKeys.FUNCTION.value not in x or x[AnnotationKeys.FUNCTION.value] is None or \
           len(x[AnnotationKeys.FUNCTION.value]) == 2 else Builder(x[AnnotationKeys.FUNCTION.value]), value


def _dirname_builder(x: dict, base_path: str = None) -> DirnameBuilder:
    """Built DirnameBuilder from an annotation based on a dirname annotation, getting the dirname which input files
    are located.
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    str
        Annotation type
    Builder
        Representation of the function to apply on the annotation value (dirname).
    re.Pattern
        Representation of a regular expression to apply on the annotation value (dirname).
    """
    func_apply, regex_apply = _get_function_and_regexp(x)

    return AnnotationTypes.DIRNAME.name, func_apply, regex_apply


def _filename_builder(x: dict, base_path: str = None) -> FilenameBuilder:
    """Built FilenameBuilder from an annotation based on a filename annotation, getting the filename of each input file.
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    str
        Annotation type
    Builder
        Representation of the function to apply on the annotation value (filename).
    re.Pattern
        Representation of a regular expression to apply on the annotation value (filename).
    """
    func_apply, regex_apply = _get_function_and_regexp(x)

    return AnnotationTypes.FILENAME.name, func_apply, regex_apply


def _get_plugin_function(mod) -> Callable:
    """Get the function from the module
    Parameters
    ----------
    mod
        Plugin module where 'run' function is imported
    Returns
    -------
    Callable
        'run' function to execute data transformation
    """
    func = None
    cls_members = inspect.getmembers(mod, inspect.isclass)
    for (_, c) in cls_members:
        if issubclass(c, Plugin) & (c is not Plugin):
            func = c().run
            break
    return func


def _get_plugin_context(mod) -> Any:
    ctxt = None
    cls_members = inspect.getmembers(mod, inspect.isclass)
    for (_, c) in cls_members:
        if issubclass(c, Context) & (c is not Context):
            ctxt = c
            break
    return ctxt


def _mapping_builder(x: dict, base_path: str) -> MappingBuilder:
    """Built MappingBuilder from an annotation based on a mapping annotation, it matches the value of the input file to
    a value that appears in the mapping file. It will return the value of one field of the mapping that has been
    indicated on the annotation.
    Parameters
    ----------
    x : dict
        Annotation.
    base_path : str
        A base path where file that is parsing is located.
    Returns
    -------
    str
        Annotation type.
    List
        Fields that has to look for in the input files.
    dict
        Schema of the mapping file, where 'key' is the value of one column (fieldMapping) in mapping file and
        'value' is the value of one column (valueMapping) in the mapping file
    """
    values: dict = {}
    mapping_files = x[AnnotationKeys.FILE_MAPPING.value]
    files = list(glob.iglob(f"{dirname(base_path)}/{mapping_files}", recursive=True))
    if len(files) == 0:
        raise FileNotFoundError(f"Unable to find '{mapping_files}' file in '{dirname(base_path)}'")
    try:
        for mapping_file in files:
            open_method = gzip.open if mapping_file.endswith('gz') else open
            with open_method(mapping_file, "rt") as fd:
                for r in csv.DictReader(fd, delimiter='\t'):
                    field = r[x[AnnotationKeys.FIELD_MAPPING.value]]
                    val = r[x[AnnotationKeys.FIELD_VALUE.value]]
                    values[field] = val
            break
    except TypeError:
        raise TypeError("Unable to parse mapping annotation")
    return AnnotationTypes.MAPPING.name, x[AnnotationKeys.FIELD_SOURCE.value], values


def _plugin_builder(x: dict, base_path: str = None) -> PluginBuilder:
    """Built PluginBuilder from an annotation based on a plugin annotation, from an internal or a customized plugin
    which data is transformed and executed thought a process.
    Parameters
    ----------
    x : dict
        Annotation
    Returns
    -------
    str
        Annotation type
    Builder
        Representation of the function to apply on the annotation value (plugin's 'run' function).
    """
    func = None
    ctxt = None
    try:
        mod = importlib.import_module(f".{x[AnnotationTypes.PLUGIN.value]}", package="openvariant.plugins")
        func = _get_plugin_function(mod)
        ctxt = _get_plugin_context(mod)
    except ModuleNotFoundError:
        try:
            files = list(glob.iglob(f"{os.getcwd()}/**/{x[AnnotationTypes.PLUGIN.value]}", recursive=True))
            if len(files) == 0:
                raise FileNotFoundError(f"Unable to find '{x[AnnotationTypes.PLUGIN.value]}' plugin in '{os.getcwd()}'")
            else:
                try:
                    for package in files:
                        spec = importlib.util.spec_from_file_location(f".{x[AnnotationTypes.PLUGIN.value]}",
                                                                      f"{package}/{x[AnnotationTypes.PLUGIN.value]}.py")
                        mod = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(mod)

                        func = _get_plugin_function(mod)
                        ctxt = _get_plugin_context(mod)
                except (ImportError, AttributeError):
                    raise ImportError(f"Unable to import 'run' on the plugin.")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(f"Unable to found '{x[AnnotationTypes.PLUGIN.value]}' plugin.")
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Unable to import the plugin: {e}")

    return AnnotationTypes.PLUGIN.name, func, ctxt


class AnnotationTypesBuilders(Enum):
    """Enum to construct every annotation type builder"""

    """Builder for static annotation"""
    STATIC = partial(_static_builder)

    """Builder for internal annotation"""
    INTERNAL = partial(_internal_builder)

    """Builder for dirname annotation"""
    DIRNAME = partial(_dirname_builder)

    """Builder for filename annotation"""
    FILENAME = partial(_filename_builder)

    """Builder for mapping annotation"""
    MAPPING = partial(_mapping_builder)

    """Builder for plugin annotation"""
    PLUGIN = partial(_plugin_builder)



"""
Count  task
====================================
A core functionality to execute count task.
"""
import functools
from multiprocessing import Pool
from os import cpu_count
from typing import Tuple, Union

from tqdm import tqdm

from openvariant.annotation.annotation import Annotation
from openvariant.find.find import find_files
from openvariant.utils.where import parse_where, skip

from openvariant.variant.variant import Variant


def _count_task(selection: Tuple[str, Annotation], group_by: str, where: str) -> Tuple[int, Union[dict, None]]:
    """Main functionality for count task"""
    where_clauses = parse_where(where)

    i = 0
    input_file, input_annotations = selection

    result = Variant(input_file, input_annotations)

    if group_by is None:
        for r in result.read():
            if skip(r, where_clauses):
                continue
            i += 1
        return i, None
    else:
        groups = {}
        for r in result.read(group_by):

            if skip(r, where_clauses):
                continue
            try:
                val = r[group_by]
                val_count = groups.get(val, 0)
                groups[val] = val_count + 1
                i += 1
            except (ValueError, KeyError):
                pass
        return i, groups


def count(base_path: str, annotation_path: str or None, group_by: str = None, where: str = None,
          cores: int = cpu_count(), quite: bool = False) -> Tuple[int, Union[None, dict]]:
    """Print on the stdout the count result.

    It'll parse the input files with its proper annotation schema, and it'll show the count result on the stdout.
    Can be grouped by a field and can be added a 'where' expression.

    Parameters
    ----------
    base_path : srt
        Base path of input files.
    annotation_path : str or None
        Path of annotation file.
    group_by : str
        Field to group the result.
    where : str
        Conditional statement.
    quite : bool
        Discard progress bar.
    cores : int
        Number of cores to parallelize the task.

    Returns
    ----------
    int
        The total number of rows.
    dict
        A schema with separate groups and the numbers of rows for each.
    """
    selection = []
    for k, a in find_files(base_path, annotation_path):
        selection += [(k, a)]

    with Pool(cores) as pool:
        groups = {}
        task = functools.partial(_count_task, group_by=group_by, where=where)
        map_method = pool.imap_unordered if len(selection) > 1 else map
        total = 0
        for c, g in tqdm(map_method(task, selection),
                         total=len(selection),
                         desc="Counting variants".rjust(40),
                         disable=(len(selection) < 2) or quite
                         ):

            # Update groups
            if g is not None:
                for k, v in g.items():
                    val = groups.get(k, 0)
                    groups[k] = val + v
            total += c

    return total, groups
